"""Ready-to-run commands for common Grafana templating scenarios."""
import datetime
import json
import logging
import textwrap
from dataclasses import dataclass
from pathlib import Path
from typing import IO

import click
from grafana_client import GrafanaApi

from grafanarmadillo.alerter import Alerter
from grafanarmadillo.bulk import BulkExporter, BulkImporter
from grafanarmadillo.dashboarder import Dashboarder
from grafanarmadillo.find import Finder, default_api_v
from grafanarmadillo.templator import (
	Templator,
	alert_dashboarduid_templator,
	make_mapping_templator,
	remove_edit_metadata_transformer,
)
from grafanarmadillo.types import GrafanaVersion
from grafanarmadillo.util import load_data


load_file_help = """Should be encoded as json. You can pass this in as a string; or as file using 'file://path/to/file'"""
auto_template_env_help = "The special value '$auto' will automatically provide a value by prepending a '$' to the keys of the grafana mapping"

mapping_help = (
	"Mapping of values to findreplace in the template. "
	+ "The type should be a mapping of environment names (like 'template' or 'production') "
	+ 'to dicts of replacements (for example, `{"region": "SouthWest"}`). '
	+ load_file_help
)
env_grafana_help = "Name of the environment in the mapping file for Grafana, found in the `--mapping` argument"
env_template_help = (
	"Name of the environment in the mapping file for the template, found in the `--mapping` argument. "
	+ auto_template_env_help
)
templator_extra_opts_help = (
	textwrap.dedent(
		"""\
		Extra options for the templator
		
		- remove_edit_metadata : remove metadata related to edits on the dashboard, such as the version and last modified times
		
		- resolve_alert_dashboarduid : resolve the dashboard referenced by an alert to a reference.
		
		"""
	) + load_file_help
)


def make_grafana(config) -> GrafanaApi:
	"""Make a GrafanaApi from the passed config."""
	if isinstance(config.get("auth"), list):
		config["auth"] = tuple(config["auth"])
	return GrafanaApi(**config)


@dataclass(frozen=True)
class TemplatorOpts:
	"""Extra options for the templator."""

	remove_edit_metadata: bool = False
	resolve_alert_dashboarduid: bool = False


def apply_template_opts(gfn: GrafanaApi, opts: TemplatorOpts, templator: Templator) -> Templator:
	"""Apply the extra templator options."""
	if opts.remove_edit_metadata:
		templator = templator.chain(Templator(make_template=remove_edit_metadata_transformer))
	if opts.resolve_alert_dashboarduid:
		templator = templator.chain(alert_dashboarduid_templator(Finder(gfn)))
	return templator


def make_templator(gfn: GrafanaApi, mapping, env_grafana, env_template, templator_extra_opts) -> Templator:
	"""Assemble the templator."""
	mapping = load_data(mapping)
	templator = make_mapping_templator(mapping, env_grafana, env_template)
	extra_opts = TemplatorOpts(**load_data(templator_extra_opts))
	templator = apply_template_opts(gfn, extra_opts, templator)
	return templator


def with_template_options(f):
	"""Add template options to a command."""
	return click.option("--templator-extra-opts", default="{}", help=templator_extra_opts_help)(
		click.option("--mapping", help=mapping_help)(
			click.option("--env-grafana", help=env_grafana_help)(
				click.option("--env-template", help=env_template_help)(f)
			)
		)
	)


@click.group()
@click.option("--cfg", "-c", help=f"Config for connecting to Grafana. {load_file_help}")
@click.option("--api-version", help="Major Grafana API version", default=default_api_v)
@click.pass_context
def grafanarmadillo(ctx, cfg, api_version):
	"""Template Grafana things."""
	ctx.ensure_object(dict)
	if cfg:
		config = load_data(cfg)
		if isinstance(config.get("auth"), list):
			config["auth"] = tuple(config["auth"])
	else:
		config = {}
	ctx.obj["cfg"] = config
	ctx.obj["api_v"] = api_version


@grafanarmadillo.group()
def dashboard():
	"""Manage Grafana dashboards."""


@dashboard.command(name="export")
@click.option("--src", help="Path to the dashboard to capture")
@click.option("--dst", help="Path to write the dashboard to", type=click.File("w"))
@with_template_options
@click.pass_context
def _export_dashboard(ctx, src, dst, mapping, env_grafana, env_template, templator_extra_opts):
	"""Capture a dashboard from Grafana."""
	gfn = make_grafana(ctx.obj["cfg"])
	templator = make_templator(gfn, mapping, env_grafana, env_template, templator_extra_opts)
	return export_dashboard(gfn, src, dst, templator, ctx.obj["api_v"])


def export_dashboard(gfn: GrafanaApi, src: str, dst: IO, templator: Templator, api_v: GrafanaVersion = default_api_v):
	"""Capture a dashboard from Grafana."""
	finder, dashboarder = Finder(gfn, api_v), Dashboarder(gfn)

	dashboard_info = finder.get_from_path(src)
	dashboard_content, _ = dashboarder.export_dashboard(dashboard_info)
	template = templator.make_template_from_dashboard(dashboard_content)

	dst.write(json.dumps(template))


@dashboard.command(name="import")
@click.option("--src", help="Path of the template", type=click.File("r"))
@click.option("--dst", help="Path to write the dashboard to")
@with_template_options
@click.pass_context
def _import_dashboard(ctx, src, dst, mapping, env_grafana, env_template, templator_extra_opts):
	"""Deploy a template to Grafana."""
	gfn = make_grafana(ctx.obj["cfg"])
	templator = make_templator(gfn, mapping, env_grafana, env_template, templator_extra_opts)
	return import_dashboard(gfn, src, dst, templator, ctx.obj["api_v"])


def import_dashboard(gfn: GrafanaApi, src: IO, dst: str, templator: Templator, api_v: GrafanaVersion = default_api_v):
	"""Deploy a template to Grafana."""
	finder, dashboarder = Finder(gfn, api_v), Dashboarder(gfn)

	template = load_data(src.read())

	dashboard_info, folder = finder.create_or_get_dashboard(dst)
	dashboard = templator.make_dashboard_from_template(dashboard_info, template)
	dashboarder.import_dashboard(dashboard, folder)


@grafanarmadillo.group()
def alert():
	"""Manage Grafana alerts."""


@alert.command(name="export")
@click.option("--src", help="Path to the alert to capture")
@click.option("--dst", help="Path to write the alert to", type=click.File("w"))
@with_template_options
@click.pass_context
def _export_alert(ctx, src, dst, mapping, env_grafana, env_template, templator_extra_opts):
	"""Capture an alert from Grafana."""
	gfn = make_grafana(ctx.obj["cfg"])
	templator = make_templator(gfn, mapping, env_grafana, env_template, templator_extra_opts)
	return export_alert(gfn, src, dst, templator, ctx.obj["api_v"])


def export_alert(gfn: GrafanaApi, src: str, dst: IO, templator: Templator, api_v: GrafanaVersion = default_api_v):
	"""Capture an alert from Grafana."""
	finder, alerter = Finder(gfn, api_v), Alerter(gfn)

	alert_info = finder.get_alert_from_path(src)
	alert, _ = alerter.export_alert(alert_info)
	template = templator.make_template_from_dashboard(alert)

	dst.write(json.dumps(template))


@alert.command(name="import")
@click.option("--src", help="Path of the template", type=click.File("r"))
@click.option("--dst", help="Path to write the dashboard to")
@with_template_options
@click.pass_context
def _import_alert(ctx, src, dst, mapping, env_grafana, env_template, templator_extra_opts):
	"""Deploy an alert from a template."""
	gfn = make_grafana(ctx.obj["cfg"])
	templator = make_templator(gfn, mapping, env_grafana, env_template, templator_extra_opts)
	return import_alert(gfn, src, dst, templator, ctx.obj["api_v"])


def import_alert(gfn: GrafanaApi, src: IO, dst: str, templator: Templator, api_v: GrafanaVersion = default_api_v):
	"""Deploy an alert from a template."""
	finder, alerter = Finder(gfn, api_v), Alerter(gfn)

	template = load_data(src.read())

	alert_info, folder_info = finder.create_or_get_alert(dst)
	alert = templator.make_dashboard_from_template(alert_info, template)
	alerter.import_alert(alert, folder_info)


@grafanarmadillo.group()
def migrate():
	"""Migrate between Grafana instances."""


@migrate.command()
@click.option(
	"--grafana-db-path",
	help="Path to the Grafana DB",
	type=click.Path(exists=True, path_type=Path),
)
@click.option(
	"--grafana-container-image",
	help="Grafana image to upgrade to",
	default="grafana/grafana:10.4.3",
)
@click.option(
	"--output-directory",
	"-o",
	help="Path to write output files",
	type=click.Path(exists=True, path_type=Path),
	default=".",
)
@click.option(
	"--grafana-extra-envvars",
	help=f"Environment variables to pass to the Grafana container used to migrate the alerts. {load_file_help}",
	default=None,
)
@click.option(
	"--grafana-migration-timeout",
	help="Timeout in seconds to wait for the Grafana docker container to apply all migrations. A large Grafana instance may take a long time",
	default=300,
	type=int,
)
@click.option(
	"--clone-db",
	help="whether to perform migrations on a clone of the DB or the DB itself",
	type=click.BOOL,
	default=True,
)
@with_template_options
@click.pass_context
def upgrade_alerting(
	ctx,
	grafana_db_path,
	grafana_container_image,
	output_directory,
	grafana_extra_envvars,
	grafana_migration_timeout,
	clone_db,
	mapping,
	env_grafana,
	env_template,
	templator_extra_opts,
):
	"""
	Migrate from Classic to Unified alerting.

	Migrations from Classic to Unified alerting is done as DB migrations.
	We clone the database and use a docker container to perform the migration.
	Note that because we use a clone of the DB, we need to supply the config with the auth for the live instance.
	"""
	from grafanarmadillo.migrate import migrate

	gfn = make_grafana(ctx.obj["cfg"])
	templator = make_templator(gfn, mapping, env_grafana, env_template, templator_extra_opts)

	cfg = ctx.obj["cfg"]
	if grafana_extra_envvars:
		grafana_extra_envvars = load_data(grafana_extra_envvars)
	else:
		grafana_extra_envvars = {}
	migrate(
		cfg,
		grafana_container_image,
		grafana_db_path,
		output_directory,
		templator,
		extra_env_vars=grafana_extra_envvars,
		timeout=datetime.timedelta(seconds=grafana_migration_timeout),
		clone_db=clone_db,
	)


@grafanarmadillo.group()
def resources():
	"""Move many resources to a Grafana."""


@resources.command("import")
@click.option(
	"--root-directory",
	help="Root directory for all resources",
	type=click.Path(exists=True, path_type=Path),
)
@with_template_options
@click.pass_context
def _import_resources(
	ctx,
	root_directory: Path,
	mapping,
	env_grafana,
	env_template,
	templator_extra_opts,
):
	"""Load exported dashboards and alerts."""
	gfn = make_grafana(ctx.obj["cfg"])
	templator = make_templator(gfn, mapping, env_grafana, env_template, templator_extra_opts)
	operator = BulkImporter(ctx.obj["cfg"], root_directory, templator=templator)
	operator.run()


@resources.command("export")
@click.option(
	"--root-directory",
	help="Root directory for all resources",
	type=click.Path(exists=True, path_type=Path),
)
@with_template_options
@click.pass_context
def _export_resources(
	ctx,
	root_directory: Path,
	mapping,
	env_grafana,
	env_template,
	templator_extra_opts,
):
	"""Export dashboards and alerts from a Grafana instance."""
	gfn = make_grafana(ctx.obj["cfg"])
	templator = make_templator(gfn, mapping, env_grafana, env_template, templator_extra_opts)
	operator = BulkExporter(ctx.obj["cfg"], root_directory, templator=templator)
	operator.run()


if __name__ == "__main__":
	logging.basicConfig(level=logging.DEBUG)
	grafanarmadillo()



"""Push and pull Grafana alerts."""
from typing import Optional, Tuple, Union

from grafana_client import GrafanaApi
from grafana_client.client import GrafanaClientError

from grafanarmadillo.types import AlertContent, AlertSearchResult, FolderSearchResult
from grafanarmadillo.util import Cache, CacheMode


class Alerter:
	"""Collection of methods for managing alert rules."""

	def __init__(self, api: GrafanaApi, disable_provenance=True, cache_mode: Union[CacheMode, Cache] = CacheMode.SESSION) -> None:
		super().__init__()
		self.api = api
		self.disable_provenance = disable_provenance
		self._cache = CacheMode.select(cache_mode)

	def import_alert(
		self, content: AlertContent, folder: FolderSearchResult
	):
		"""Import an alert into Grafana."""
		content = content.copy()
		# set the folder in case it isn't, which would happen if the metadata was scrubbed from the alert content
		content["folderUID"] = folder["uid"]
		content.pop("id", None)

		try:
			if "uid" in content:
				exists = self.api.alertingprovisioning.get_alertrule(content["uid"])
			else:
				exists = None
		except GrafanaClientError as e:
			if e.status_code == 404:
				exists = None
			else:
				raise

		if exists:
			self.api.alertingprovisioning.update_alertrule(content["uid"], content, disable_provenance=self.disable_provenance)
		else:
			self.api.alertingprovisioning.create_alertrule(content, disable_provenance=self.disable_provenance)
		self._cache.unset("list_alerts")

	def export_alert(
		self, alert: AlertSearchResult
	) -> Tuple[AlertContent, Optional[FolderSearchResult]]:
		"""Export an alert from Grafana and its folder information too."""
		alert_content = self.api.alertingprovisioning.get_alertrule(alert["uid"])

		folder = self.api.folder.get_folder(alert_content["folderUID"])

		return alert_content, folder



"""Pieces for templating multiple dashboards at once."""
from __future__ import annotations

import json
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from pathlib import Path
from typing import Iterable, List, Optional, Type, Union

import urllib3
from grafana_client import GrafanaApi

from grafanarmadillo.alerter import Alerter
from grafanarmadillo.dashboarder import Dashboarder
from grafanarmadillo.find import Finder
from grafanarmadillo.templator import Templator
from grafanarmadillo.types import PathLike
from grafanarmadillo.util import CacheMode, resolve_object_to_filepath


class Store(ABC):
	"""A destination or source for items."""

	@abstractmethod
	def read_alert(self, name: PathLike):
		"""Read an alert from this store."""

	@abstractmethod
	def read_dashboard(self, name: PathLike):
		"""Read a dashboard from this store."""

	@abstractmethod
	def write_alert(self, name: PathLike, alert):
		"""Write an alert to this store."""

	@abstractmethod
	def write_dashboard(self, name: PathLike, dashboard):
		"""Write an alert to this store."""


@dataclass
class FileStore(Store):
	"""
	Store and retrieve Grafana objects in the filesystem.

	Objects will be stored under the same path as in Grafana.
	For example, a dashboard titled "MyDashboard" in a folder titled "MyFolder"
	will appear at `{root}/MyFolder/MyDashboard.json`.

	You can customise the path these files will be stored under.
	Override the method `resolve_object_to_filepath`.
	"""

	root: Path
	json_encoder: Type[json.JSONEncoder] = json.JSONEncoder
	json_decoder: Type[json.JSONDecoder] = json.JSONDecoder

	TOK_DASHBOARD = "dashboard"
	TOK_ALERT = "alert"

	@staticmethod
	def _read(file: Path, codec: Type[json.JSONDecoder]) -> dict:
		with file.with_suffix(".json").open(mode="r", encoding="utf-8") as f:
			return json.load(f, cls=codec)

	@staticmethod
	def _write(file: Path, content: dict, codec: Type[json.JSONEncoder]):
		file.parent.mkdir(exist_ok=True)
		with file.with_suffix(".json").open(mode="w", encoding="utf-8") as f:
			json.dump(content, f, cls=codec)

	def resolve_object_to_filepath(self, name: PathLike, type_: str):
		"""
		Find the file on disk that contains the object.

		@param name:
		@param type_: The kind of object to find, one of "dashboard" or "alert".
		"""
		return resolve_object_to_filepath(self.root, name)

	def read_alert(self, name):
		"""Read an alert from this store."""
		return self._read(self.resolve_object_to_filepath(name, self.TOK_ALERT), self.json_decoder)

	def read_dashboard(self, name):
		"""Read a dashboard from this store."""
		return self._read(self.resolve_object_to_filepath(name, self.TOK_DASHBOARD), self.json_decoder)

	def write_alert(self, name, alert):
		"""Write an alert to this store."""
		return self._write(self.resolve_object_to_filepath(name, self.TOK_ALERT), alert, self.json_encoder)

	def write_dashboard(self, name, dashboard):
		"""Write an alert to this store."""
		return self._write(self.resolve_object_to_filepath(name, self.TOK_DASHBOARD), dashboard, self.json_encoder)


class GrafanaStore(Store):
	"""Store and retrieve objects from a Grafana instance."""

	def __init__(self, gfn: GrafanaApi):
		self.gfn = gfn
		self.finder = Finder(gfn, cache_mode=CacheMode.SESSION)

	def read_alert(self, name):
		"""Read an alert from this store."""
		alerter = Alerter(self.gfn)
		alert_info, _ = self.finder.create_or_get_alert(name)
		alert, _ = alerter.export_alert(alert_info)
		return alert_info

	def read_dashboard(self, name):
		"""Read a dashboard from this store."""
		dashboarder = Dashboarder(self.gfn)
		dashboard_info, _ = self.finder.create_or_get_dashboard(name)
		dashboard_content, _ = dashboarder.export_dashboard(dashboard_info)
		return dashboard_content

	def write_alert(self, name, alert):
		"""Write an alert to this store."""
		alerter = Alerter(self.gfn)
		alert_info, folder_info = self.finder.create_or_get_alert(name)
		alerter.import_alert(alert, folder_info)

	def write_dashboard(self, name, dashboard):
		"""Write an alert to this store."""
		dashboarder = Dashboarder(self.gfn)
		dashboard_info, folder = self.finder.create_or_get_dashboard(name)
		dashboarder.import_dashboard(dashboard, folder)


@dataclass
class URLStore(Store):
	"""Store and retrieve objects from remote URLs, such as https://grafana.com/grafana/dashboards/ or GitHub."""

	def read_url(self, url: str):
		"""GET a URL."""
		resp = urllib3.request("GET", url)
		return resp.json()

	def read_alert(self, name):
		"""Read an alert from this store."""
		if not isinstance(name, str):
			raise TypeError(f"URLStore can only read remote URLs, not {type(name)}")
		return self.read_url(name)

	def read_dashboard(self, name):
		"""Read a dashboard from this store."""
		if not isinstance(name, str):
			raise TypeError(f"URLStore can only read remote URLs, not {type(name)}")
		return self.read_url(name)

	def write_alert(self, name, alert):
		"""Write an alert to this store."""
		raise NotImplementedError("URLStore cannot write")

	def write_dashboard(self, name, dashboard):
		"""Write an alert to this store."""
		raise NotImplementedError("URLStore cannot write")


@dataclass
class Alert:
	"""Flowable request for an Alert."""

	name_obj: PathLike
	name_tmpl: PathLike
	templator: Templator


@dataclass
class Dashboard:
	"""Flowable request for a Dashboard."""

	name_obj: PathLike
	name_tmpl: PathLike
	templator: Templator


Flowable = Union[Alert, Dashboard]


class FlowException(Exception):
	"""Wrapped Exception of running a Flow."""

	def __init__(self, item: Flowable, cause: Optional[BaseException] = None):
		self.item = item
		if cause:
			self.__cause__ = cause


@dataclass
class FlowResult:
	"""Result of running a Flow."""

	successes: List[Flowable]
	failures: List[FlowException]

	def raise_first(self):
		"""Raise the first exception, if present."""
		if self.failures:
			raise self.failures[0]

	def ensure_success(self) -> FlowResult:
		"""Inline raise the first exception, if present."""
		self.raise_first()
		return self


@dataclass
class Flow:
	"""A collection of templating actions to do."""

	store_obj: Store
	store_tmpl: Store
	flows: List[Flowable] = field(default=list)

	def append(self, flow: Flowable):
		"""Add a Flowable request to this Flow."""
		self.flows.append(flow)

	def extend(self, flows: Iterable[Flowable]):
		"""Add several Flowable requests to this Flow."""
		self.flows.extend(flows)

	def obj_to_tmpl(self) -> FlowResult:
		"""Import from the source to the destination."""
		return self.run(obj_to_tmpl=True)

	def tmpl_to_obj(self) -> FlowResult:
		"""Export from the destination to the source."""
		return self.run(obj_to_tmpl=False)

	def run(self, obj_to_tmpl: bool) -> FlowResult:
		"""Run the flow."""
		failures = []
		successes = []

		for item in self.flows:
			try:
				if isinstance(item, Alert):
					if obj_to_tmpl:
						obj = self.store_obj.read_alert(item.name_obj)
						tmpl = item.templator.make_template_from_dashboard(obj)
						self.store_tmpl.write_alert(item.name_tmpl, tmpl)
					else:
						tmpl = self.store_tmpl.read_alert(item.name_tmpl)
						info = self.store_obj.read_alert(item.name_obj)
						obj = item.templator.make_dashboard_from_template(info, tmpl)
						self.store_obj.write_alert(item.name_obj, obj)
				elif isinstance(item, Dashboard):
					if obj_to_tmpl:
						obj = self.store_obj.read_dashboard(item.name_obj)
						tmpl = item.templator.make_template_from_dashboard(obj)
						self.store_tmpl.write_dashboard(item.name_tmpl, tmpl)
					else:
						tmpl = self.store_tmpl.read_dashboard(item.name_tmpl)
						info = self.store_obj.read_dashboard(item.name_obj)
						obj = item.templator.make_dashboard_from_template(info, tmpl)
						self.store_obj.write_dashboard(item.name_obj, obj)
				else:
					raise TypeError(
						f"Invalid flow, expected one of {Alert.__name__}, {Dashboard.__name__}, received {item.__class__.__name__}")

				successes.append(item)
			except Exception as e:
				failures.append(FlowException(item, e))

		return FlowResult(successes, failures)



"""Type hints for Grafana interaction."""
from dataclasses import dataclass
from pathlib import Path
from typing import List, NewType, Optional, TypedDict, Union


GrafanaVersion = NewType("UID", int)
UID = NewType("UID", str)


class AnySearchResult(TypedDict):
	"""Metadata for both Grafana dashboards and alerts."""

	uid: UID
	title: str
	folderUid: Optional[UID]


class DashboardSearchResult(TypedDict):
	"""Relevant keys returned by a Grafana search."""

	id: int
	uid: UID
	title: str
	folderId: Optional[int]
	folderUid: Optional[UID]
	folderTitle: Optional[str]


class AlertSearchResult(TypedDict):
	"""Relevant keys returned for an alert."""

	id: int
	uid: UID
	orgID: int
	folderUID: UID
	ruleGroup: str
	title: str


DashboardContent = NewType("DashboardContent", dict)
AlertContent = NewType("AlertContent", dict)
AnyContent = Union[DashboardContent, AlertContent]
DashboardMeta = NewType("DashboardMeta", dict)

OrgMeta = NewType("OrgMeta", dict)


class Dashboard(TypedDict):
	"""Relevant keys returned by GET of a Grafana dashboard."""

	meta: DashboardMeta
	dashboard: DashboardContent


DashboardPanel = NewType("DashboardPanel", dict)

FolderSearchResult = NewType("Folder", dict)


@dataclass
class GrafanaPath:
	"""Path of an object in Grafana."""

	name: str
	folder: str
	org: Optional[str] = None


PathLike = Union[GrafanaPath, str, Path, List[str]]



"""Push and pull Grafana dashboards."""
from typing import Optional, Tuple

from grafana_client import GrafanaApi

from grafanarmadillo.types import (
	DashboardContent,
	DashboardSearchResult,
	FolderSearchResult,
)
from grafanarmadillo.util import project_dashboard_identity


class Dashboarder:
	"""Collection of methods for managing dashboards."""

	def __init__(self, api: GrafanaApi) -> None:
		super().__init__()
		self.api = api

	def get_dashboard_content(self, dashboard: DashboardSearchResult) -> DashboardContent:
		"""Get the contents of a Grafana dashboard."""
		return self.api.dashboard.get_dashboard(dashboard["uid"])["dashboard"]

	def set_dashboard_content(
		self, dashboard: DashboardSearchResult, content: DashboardContent
	):
		"""
		Set the content of a Grafana dashboard.
		
		This explicitly leaves out the identity information.
		That allows you to graft the contents of a dashboard into another
		"""
		new_dashboard = dashboard.copy()
		new_content = content.copy()

		new_content.update(project_dashboard_identity(new_dashboard))

		new_dashboard.update({"dashboard": new_content, "overwrite": True})

		return self.api.dashboard.update_dashboard(new_dashboard)

	def import_dashboard(
		self, content: DashboardContent, folder: Optional[FolderSearchResult] = None
	):
		"""Import a dashboard into Grafana, optionally into a folder."""
		content.pop("id", None)
		new_dashboard = {"dashboard": content, "overwrite": True}
		if folder:
			new_dashboard.update({"folderUid": folder["uid"], "folderId": folder["id"]})

		return self.api.dashboard.update_dashboard(new_dashboard)

	def export_dashboard(
		self, dashboard: DashboardSearchResult
	) -> Tuple[DashboardContent, Optional[FolderSearchResult]]:
		"""Export a dashboard from grafana, with its folder information if applicable."""
		result = self.api.dashboard.get_dashboard(dashboard["uid"])
		meta, dashboard = result["meta"], result["dashboard"]
		if meta["folderUid"]:
			folder = self.api.folder.get_folder(meta["folderUid"])
		else:
			folder = None

		return dashboard, folder



"""Migrate from Classic to Unified alerting."""
import contextlib
import datetime
import logging
import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import Dict

import docker
import requests
from docker.models.containers import Container

from grafanarmadillo.bulk import BulkExporter
from grafanarmadillo.templator import Templator


l = logging.getLogger(__name__)


@dataclass
class DockerContainer:
	"""Handle on a docker container."""

	container: Container
	image: str
	host_port: int

	@property
	def status(self):
		"""Get the status of the docker container."""
		return self.container.status


def start_container(image_name, volume_path: Path, environment_vars: Dict[str, str]):
	"""Start a container."""
	client = docker.from_env()
	volumes = {str(volume_path): {"bind": "/var/lib/grafana/grafana.db", "mode": "rw"}}
	container = client.containers.run(
		image_name,
		detach=True,
		ports={"3000/tcp": 0},
		volumes=volumes,
		environment=environment_vars,
	)
	container.reload()
	host_port = container.attrs["NetworkSettings"]["Ports"]["3000/tcp"][0]["HostPort"]
	return DockerContainer(container, image_name, int(host_port))


def read_container_logs(container: DockerContainer):
	"""Read logs from the docker container."""
	return container.container.logs().decode("utf-8")


def stop_container(container: DockerContainer):
	"""Stop container."""
	container.container.stop()


def exec_in_container(container: DockerContainer, command: str):
	"""Execute a command in a running docker container."""
	result = container.container.exec_run(command)
	return result.output.decode("utf-8").strip()


@contextlib.contextmanager
def with_container(image_name, volume_path: Path, environment_vars: Dict[str, str]):
	"""Context manager for Grafana docker container."""
	container = start_container(image_name, volume_path, environment_vars)
	try:
		yield container
	finally:
		stop_container(container)


DEFAULT_TIMEOUT = datetime.timedelta(seconds=300)


def _wait_until_ready(
	container: DockerContainer,
	timeout: datetime.timedelta = DEFAULT_TIMEOUT,
):
	"""Wait until container's readiness check passes."""
	start = datetime.datetime.now()
	end = start + timeout

	l.debug(f"waiting for container to be ready start={start}, end={end}, timeout={timeout}")
	while True:
		if datetime.datetime.now() > end:
			raise RuntimeError(
				f"Could not connect to container in {timeout} logs={read_container_logs(container)}"
			)
		try:
			if requests.get(f"http://localhost:{container.host_port}/api/health").ok:
				break
		except (
			ConnectionError,
			requests.exceptions.ConnectionError,
			requests.exceptions.HTTPError,
		):
			pass


def migrate(
	cfg: dict,
	grafana_image: str,
	grafana_db: Path,
	output_directory: Path,
	templator: Templator,
	extra_env_vars: Dict[str, str] = None,
	grafana_uid: int = 472,
	timeout: datetime.timedelta = DEFAULT_TIMEOUT,
	clone_db: bool = True
) -> None:
	"""Migrate from classic to Unified alerting."""
	extra_env_vars = extra_env_vars or {}

	if clone_db:
		new_db = grafana_db.with_name("migrated").absolute()
		l.debug(f"cloning db from={grafana_image} to={new_db}")
		shutil.copyfile(grafana_db, new_db)
		if not new_db.stat().st_uid == grafana_uid:
			try:
				import subprocess

				subprocess.run(["sudo", "chown", str(grafana_uid), new_db.as_posix()])
			except PermissionError:
				l.warning(f"Could not change owner of Grafana DB. expected={grafana_uid} actual={new_db.stat().st_uid} permissions={oct(new_db.stat().st_mode)}")
	else:
		new_db = grafana_db

	l.debug("begin migrating")
	with with_container(grafana_image, new_db, extra_env_vars) as container:
		if container.status != "running":
			raise RuntimeError(f"Could not start Grafana container {container=}")

		l.info("wait for migrations to apply")
		_wait_until_ready(container, timeout=timeout)
		l.info("migrations applied")

		l.info("export dashboards")
		exporter = BulkExporter(
			{**cfg, **{
				"host": "localhost",
				"port": container.host_port,
			}},
			output_directory,
			templator=templator,
		)

		exporter.run()
		l.info("export dashboards complete")



"""Helpers and generic functions."""
from __future__ import annotations

import json
import logging
from enum import Enum
from pathlib import Path
from typing import Callable, Dict, List, TypeVar, Union

from grafanarmadillo.paths import PathCodec
from grafanarmadillo.types import (
	DashboardContent,
	DashboardSearchResult,
	GrafanaPath,
	PathLike,
)


T = TypeVar("T")
A = TypeVar("A")
JSON = TypeVar("JSON", bound=Union[dict, list, str, int, float, bool, None])


def flat_map(f, xs):
	"""
	Flatmap: Map on a list and then merge the results.
	
	>>> and_reversed = lambda s: [s,s[::-1]]

	>>> flat_map(and_reversed, [])
	[]

	>>> flat_map(and_reversed, ['hi'])
	['hi', 'ih']

	>>> flat_map(and_reversed, ['hi', 'hello'])
	['hi', 'ih', 'hello', 'olleh']

	"""
	ys = []
	for x in xs:
		ys.extend(f(x))
	return ys


def exactly_one(items: List[A], message: str = None) -> A:
	"""
	Throws if list does not contain exactly 1 item.
	
	>>> exactly_one([1])
	1

	>>> exactly_one([1,2])
	Traceback (most recent call last):
	ValueError: expected exactly 1 item, found=2 message=None

	>>> exactly_one([])
	Traceback (most recent call last):
	ValueError: expected exactly 1 item, found=0 message=None

	"""
	if len(items) == 1:
		return items[0]
	else:
		raise ValueError(f"expected exactly 1 item, found={len(items)} {message=}")


def project_dict(d: Dict, keys: set, inverse: bool = False) -> Dict:
	"""
	Select the given fields from a dictionary.
	
	>>> project_dict({'a': 1, 'b': 2}, {'a'})
	{'a': 1}

	>>> project_dict({'a': 1, 'b': 2}, {'a'}, inverse=True)
	{'b': 2}
	"""
	return {k: v for k, v in d.items() if inverse ^ (k in keys)}


dashboard_meta_fields = {"id", "uid", "title"}
alert_rule_meta_fields = {"id", "uid", "title", "orgID", "folderUID"}


def project_dashboard_identity(
	dashboardlike: Union[DashboardSearchResult, DashboardContent]
) -> Dict:
	"""Project only the fields of a dashboard which are used for determining identity."""
	return project_dict(dashboardlike, dashboard_meta_fields)


def erase_dashboard_identity(
	dashboardlike: Union[DashboardSearchResult, DashboardContent]
) -> Dict:
	"""Delete the fields of a dashboard which are used for determining identity."""
	return project_dict(dashboardlike, dashboard_meta_fields, inverse=True)


def erase_alert_rule_identity(
	alertlike
) -> Dict:
	"""Delete the fields of an alert_rule which are used for determining identity."""
	return project_dict(alertlike, alert_rule_meta_fields, inverse=True)


def map_json_strings(f: Callable[[str], str], obj: JSON) -> JSON:
	"""
	Transform all strings in an object made of JSON primitives.

	>>> f = lambda s: s.upper()
	>>> map_json_strings(f, 's')
	'S'
	>>> map_json_strings(f, 1)
	1
	>>> map_json_strings(f, ['s'])
	['S']
	>>> map_json_strings(f, ['s', 1])
	['S', 1]
	>>> map_json_strings(f, {'a': 's'})
	{'a': 'S'}
	>>> map_json_strings(f, {'a': ['s', 1]})
	{'a': ['S', 1]}
	"""
	if isinstance(obj, dict):
		return {k: map_json_strings(f, v) for k, v in obj.items()}
	elif isinstance(obj, list):
		return [map_json_strings(f, i) for i in obj]
	elif isinstance(obj, str):
		return f(obj)
	else:
		return obj


def resolve_object_to_filepath(base_path: Path, name: PathLike):
	"""Transform the "/folder/object" format to the path on disk that contains the template."""
	path = PathCodec.encode_grafana(PathCodec.try_parse(name))
	template_path = (base_path / path).with_suffix(".json")
	return template_path


def resolve_filepath_to_object(base_path: Path, path: Path) -> GrafanaPath:
	"""Extract the "/folder/object" format from the file on disk that contains the template."""
	template_path = path.relative_to(base_path).with_suffix("")
	name = PathCodec.try_parse(PathCodec.decode(template_path))
	return name


def load_data(data_str: str):
	"""Attempt to load data."""
	_file_uri_prefix = "file://"
	if data_str.startswith(_file_uri_prefix):
		filename = Path(data_str.split(_file_uri_prefix)[1])
		with filename.open(mode="r", encoding="utf-8") as data_file:
			return json.load(data_file)
	else:
		return json.loads(data_str)


def write_to_file(out_path: Path, obj: dict):
	"""Write an object to file as JSON."""
	out_path.parent.mkdir(parents=True, exist_ok=True)
	with out_path.open(mode="w+", encoding="utf-8") as f:
		json.dump(obj, f, ensure_ascii=False, indent="\t")


def read_from_file(file_path: Path) -> dict:
	"""Read JSON from a file."""
	with file_path.open(mode="r", encoding="utf-8") as f:
		return json.load(f)


class CacheMode(Enum):
	"""
	Caching mode for interacting with Grafana.

	None: no caching
	Session: lifetime of the Finder object
	Global: all Finders share the same cache

	You can disable caching globally by setting `grafanarmadillo.util.global_cache = grafanarmadillo.util.NoneCache()`
	"""

	NONE = "NONE"
	SESSION = "SESSION"
	GLOBAL = "GLOBAL"

	@staticmethod
	def select(cache_mode: Union[CacheMode, Cache]) -> Cache:
		"""Create or use a cache. Pass a cache to reuse it."""
		if isinstance(cache_mode, Cache):
			return cache_mode

		if cache_mode == CacheMode.GLOBAL:
			return global_cache
		elif cache_mode == CacheMode.SESSION:
			return Cache()
		else:
			return NoneCache()


l_c = logging.getLogger(f"{__name__}.cache")


class Cache:
	"""Cache values."""

	def __init__(self):
		self.cache = {}

	def get(self, k):
		"""Get a cached value, if it exists."""
		return self.cache.get(k, None)

	def set(self, k, v):
		"""Set a cached value."""
		self.cache[k] = v

	def unset(self, k):
		"""Unset a cached value."""
		self.cache.pop(k, None)

	def unset_method(self, k_start):
		"""Unset all keys whose first subkey (the method name) matches."""
		cull = set()

		for k in self.cache.keys():
			if k[0] == k_start:
				cull.add(k)
		for k in cull:
			self.unset(k)

	def getor(self, k, f: Callable[[], T]) -> T:
		"""Get a cached item or generate it."""
		if v := self.get(k):
			l_c.debug(f"cache hit {k}")
			return v
		l_c.debug(f"cache miss {k}")
		v = f()
		self.set(k, v)
		return v


global_cache = Cache()


class NoneCache(Cache):
	"""A Cache-interface-compatible which never caches."""

	def get(self, k):
		"""Never caches a value."""
		return None

	def set(self, k, v):
		"""Never caches a value."""
		return

	def unset(self, k):
		"""No keys are ever set."""
		return

	def getor(self, k, f: Callable[[], T]) -> T:
		"""Always generate the cached item."""
		return f()
